(ns rlm.image-utils
  (:import [javax.swing JFrame]
	   [java.awt Color BorderLayout]
	   [ij ImagePlus IJ]
	   [java.lang Math]
	   [java.awt Graphics2D Panel]
	   [ij Macro]
	   [java.io BufferedReader InputStreamReader]
	   [java.awt.image BufferedImage])
  (:use [clojure.contrib
	 [math]
	 [def]]))


(defmulti frame-hash-multi  class)

(defmethod frame-hash-multi ImagePlus
  [image+]
    (with-meta
      (let [buf (.. image+ getBufferedImage)
	    color (.getColorModel buf)]
	(apply hash-map 
	       (interleave 
		(doall (for [x (range (.getWidth image+)) y (range (.getHeight image+))]
			 (vector x y)))
		(doall (for [x (range (.getWidth image+)) y (range (.getHeight image+))]
			 (let [data (.getRGB buf x y)] 
			   (hash-map :r (bit-shift-right (bit-and 0xff0000 data) 16)
				     :g (bit-shift-right (bit-and 0x00ff00 data) 8)
				     :b (bit-and 0x0000ff data))))))))
      {:width  (.getWidth image+) :height (.getHeight image+)}))

(defmethod frame-hash-multi String
  [image-name]
  (let [image+ (ImagePlus. image-name)]
    (frame-hash-multi image+)))

(defn frame-hash
  "yields a convienent representation for the pixles in an image.
   Because of the size of the structvre generated, this must only be used
   in a transient way so that java can do it's garbage collection."
  [something]
  (frame-hash-multi something))


(def white {:r 255, :g 255, :b 255})
(def black {:r 0,   :g 0,   :b 0})

(defn rgb-euclidian
  [{r1 :r g1 :g b1 :b} {r2 :r g2 :g b2 :b} ]
  (expt (+ (expt (- r1 r2) 2)
	  (expt (- g1 g2) 2)
	  (expt (- b1 b2) 2)) 0.5))

(defn b&w
  "turn everything strictly black or white"
  [window]
  (with-meta
  (zipmap
   (keys window)
   (map (fn  [rgb] 
	  (if (> (rgb-euclidian rgb white) (rgb-euclidian rgb black))
	    black white))
	(vals window))) (meta window)))

(defn frame-hash->bufferedImage
  [frame-hash]
    (let [data  (meta frame-hash)
	image (BufferedImage. (:width data) (:height data) BufferedImage/TYPE_INT_BGR)]
	
	(doall (for [element frame-hash] 
	  (let [coord  (key element) 
		rgb   (val element)
		packed-RGB 
		(+ (bit-shift-left (:r rgb) 16)
		   (bit-shift-left (:g rgb) 8)
		                   (:b rgb))]
	    (.setRGB image (first coord) (last coord) packed-RGB))))
	image))
  
(defmulti  display "Creates a JFrame and displays a buffered image"  class)

(defn- makePanel [image]
  (proxy [Panel] [] (paint [g]  (.drawImage g image 0 0 nil))))

(defmethod display 
  BufferedImage  [image] 
  (let [panel (makePanel image)
	frame (JFrame. "Oh Yeah!")]
    (.add frame panel) 
    (.pack frame) 
    (.setVisible frame true ) 
    (.setSize frame(.getWidth image) (.getHeight image))))
 
(defmethod display
  ImagePlus [image]
  (display (.getBufferedImage image)))

(defmethod display
  clojure.lang.PersistentHashMap [frame-hash]
  (display (frame-hash->bufferedImage frame-hash)))

(defmethod display
    clojure.lang.PersistentArrayMap [frame-hash]
    (display (frame-hash->bufferedImage frame-hash)))

(defn rotate [degrees #^ImagePlus image]
  (.rotate (.getChannelProcessor image) degrees)
  image)


(defn-memo width [pic]
  (inc (first (last (sort (fn [[x0 y0] [x1 y1]] (> x1 x0)) (keys pic))))))

(defn-memo height [pic]
    (inc (last (last (sort (fn [[x0 y0] [x1 y1]] (> y1 y0)) (keys pic))))))

